from linux_router import LinuxRouter
from mininet.topo import Topo
from topologies.custom_topo import CustomTopo


class FourRoutersFourHosts(CustomTopo):
    def build(self):
        # we pre defined count of hosts for cpu limitations
        self.host_count = 8
        # Adding routers
        R1 = self.addNode('r1', cls=LinuxRouter, ip='10.1.0.1/24')
        R2 = self.addNode('r2', cls=LinuxRouter, ip='10.2.0.1/24')
        R3 = self.addNode('r3', cls=LinuxRouter, ip='10.3.0.1/24')
        R4 = self.addNode('r4', cls=LinuxRouter, ip='10.4.0.1/24')

        # Adding switches
        S1 = self.addSwitch('s1')
        S2 = self.addSwitch('s2')
        S3 = self.addSwitch('s3')
        S4 = self.addSwitch('s4')

        # Connecting switches to routers
        self.addLink(S1, R1, intfName2='r1-eth0', params2={'ip': '10.1.0.1/24'})
        self.addLink(S2, R2, intfName2='r2-eth0', params2={'ip': '10.2.0.1/24'})
        self.addLink(S3, R3, intfName2='r3-eth0', params2={'ip': '10.3.0.1/24'})
        self.addLink(S4, R4, intfName2='r4-eth0', params2={'ip': '10.4.0.1/24'})

        # Connecting routers to each other
        self.addLink(R1, R2, intfName1='r1-eth1', intfName2='r2-eth1', params1={'ip': '10.112.0.11/24'},
                     params2={'ip': '10.112.0.21/24'})
        self.addLink(R1, R3, intfName1='r1-eth2', intfName2='r3-eth1', params1={'ip': '10.113.0.12/24'},
                     params2={'ip': '10.113.0.31/24'})
        self.addLink(R2, R4, intfName1='r2-eth2', intfName2='r4-eth1', params1={'ip': '10.124.0.22/24'},
                     params2={'ip': '10.124.0.41/24'})
        self.addLink(R3, R4, intfName1='r3-eth2', intfName2='r4-eth2', params1={'ip': '10.134.0.32/24'},
                     params2={'ip': '10.134.0.42/24'})

        h1 = self.addHost('h1', ip='10.1.0.101/24', defaultRoute='via 10.1.0.1')
        h2 = self.addHost('h2', ip='10.2.0.101/24', defaultRoute='via 10.2.0.1')
        h3 = self.addHost('h3', ip='10.3.0.101/24', defaultRoute='via 10.3.0.1')
        h4 = self.addHost('h4', ip='10.4.0.101/24', defaultRoute='via 10.4.0.1')

        for h, s in [(h1, S1), (h2, S2), (h3, S3), (h4, S4)]:
            self.addLink(h, s)

    def get_routings(self):
        return {
            "r1": {
                "default": [
                    ("10.2.0.0/24", "10.112.0.21", "r1-eth1", 10),
                    ("10.3.0.0/24", "10.113.0.31", "r1-eth2", 10),
                    ("10.4.0.0/24", "10.112.0.21", "r1-eth1", 10),
                    ("10.4.0.0/24", "10.113.0.31", "r1-eth2", 20),  # "fast-rerouting", in case r2 is disconnected
                    ("10.124.0.0/24", "10.112.0.21", "r1-eth1", 10),
                    ("10.134.0.0/24", "10.113.0.31", "r1-eth2", 10)
                ],
                "incoming_eth1": [
                    ("10.4.0.0/24", "10.113.0.31", "r1-eth2", 10),
                ]
            },
            "r2": {
                "default": [
                    ("10.1.0.0/24", "10.112.0.11", "r2-eth1", 10),
                    ("10.3.0.0/24", "10.112.0.11", "r2-eth1", 10),
                    ("10.4.0.0/24", "10.124.0.41", "r2-eth2", 10),
                    ("10.4.0.0/24", "10.112.0.11", "r2-eth1", 20),  # "fast-rerouting" in case connection to r4 is lost
                    ("10.134.0.0/24", "10.124.0.41", "r2-eth2", 10),
                    ("10.113.0.0/24", "10.112.0.11", "r2-eth1", 10),
                ]
            },
            "r3": {
                "default": [
                    ("10.1.0.0/24", "10.113.0.12", "r3-eth1", 10),
                    ("10.2.0.0/24", "10.113.0.12", "r3-eth1", 10),
                    ("10.4.0.0/24", "10.134.0.42", "r3-eth2", 10),
                    ("10.124.0.0/24", "10.134.0.42", "r3-eth2", 10),
                    ("10.112.0.0/24", "10.113.0.12", "r3-eth1", 10)
                ],
            },
            "r4": {
                "default": [
                    ("10.1.0.0/24", "10.124.0.22", "r4-eth1", 10),
                    ("10.1.0.0/24", "10.134.0.32", "r4-eth2", 20),
                    ("10.2.0.0/24", "10.124.0.22", "r4-eth1", 10),
                    ("10.3.0.0/24", "10.134.0.32", "r4-eth2", 10),
                    ("10.112.0.0/24", "10.124.0.22", "r4-eth1", 10),
                    ("10.113.0.0/24", "10.134.0.32", "r4-eth2", 10)
                ]
            }
        }

    def get_policies(self):
        return {
            "r1": {
                "rules": [
                    {
                        "table": "incoming_eth1",
                        "inport": "r1-eth1",
                        "to": ["10.4.0.0/24"]
                    },
                    {
                        "table": "incoming_eth2",
                        "inport": "r1-eth2",
                        "to": ["10.3.0.0/24"]
                    }
                ]
            },
            "r2": {
                "rules": [
                    {
                        "table": "incoming_eth1",
                        "inport": "r2-eth1",
                        "to": ["10.3.0.0/24"]
                    }
                ]
            }

        }

    def get_tests(self):
        return {
            "tcp_bandwidth_intermediate": {
                "pre_execution": {
                    "use_pre_defined_function": False,
                    "command": lambda net: net.iperf((net['h1'], net['h4'])),
                },
                "source": "h1",
                "destination": "h4",
                "destination_ip": "10.4.0.101",
                "execute": {
                    "use_pre_defined_function": True,
                    "separate_definitions": True,
                    "command_pre": ("measure_bandwidth", (['h1', 'h4'], '10.4.0.101', 30, 0.5, "tcp_pre_failure", "TCP Bandwidth before failure", "tcp", 100, [0, self.limit+self.limit*0.1])),
                    "command_post": ("measure_bandwidth", (['h1', 'h4'], '10.4.0.101', 30, 0.5, "tcp_post_failure", "TCP Bandwidth after failure", "tcp", 100, [0, self.limit+self.limit*0.1]))
                },
                "failures": [
                    {
                        "type": "intermediate",
                        "execute": {
                            "use_pre_defined_function": True,
                            "command": (
                            "connection_shutdown", (['r2', 'r4'], ['Router 2', 'Router 4'], ['r2-eth2', 'r4-eth1']))
                        }
                    }
                ]
            },
            "udp_bandwidth_intermediate": {
                "pre_execution": {
                    "use_pre_defined_function": False,
                    "command": lambda net: net.iperf((net['h1'], net['h4'])),
                },
                "source": "h1",
                "destination": "h4",
                "destination_ip": "10.4.0.101",
                "execute": {
                    "use_pre_defined_function": True,
                    "separate_definitions": True,
                    "command_pre": ("measure_bandwidth", (
                    ['h1', 'h4'], '10.4.0.101', 30, 0.5, "tcp_pre_failure", "UDP Bandwidth before failure", "tcp", 100,
                    [0, self.limit + self.limit * 0.1])),
                    "command_post": ("measure_bandwidth", (
                    ['h1', 'h4'], '10.4.0.101', 30, 0.5, "tcp_post_failure", "UDP Bandwidth after failure", "tcp", 100,
                    [0, self.limit + self.limit * 0.1]))
                },
                "failures": [
                    {
                        "type": "intermediate",
                        "execute": {
                            "use_pre_defined_function": True,
                            "command": (
                                "connection_shutdown", (['r2', 'r4'], ['Router 2', 'Router 4'], ['r2-eth2', 'r4-eth1']))
                        }
                    }
                ]
            },
            "bandwidth_link_usage_intermediate": {
                "pre_execution": {
                    "use_pre_defined_function": False,
                    "command": lambda net: net.iperf((net['h1'], net['h4'])),
                },
                "source": "h1",
                "destination": "h4",
                "destination_ip": "10.4.0.101",
                "execute": {
                    "use_pre_defined_function": True,
                    "separate_definitions": True,
                    "command_pre": ("measure_link_usage_bandwidth", (
                        {"client": "h1", "server": "h4", "server_ip": "10.4.0.101"}, {"client": "h2", "server": "h1", "server_ip": "10.1.0.101"}, 30, 1, "bandwidth_link_usage_pre_failure", "TCP Bandwidth before failure", "tcp", 100,
                    [0, 130])),
                    "command_post": ("measure_link_usage_bandwidth", (
                        {"client": "h1", "server": "h4", "server_ip": "10.4.0.101"}, {"client": "h2", "server": "h1", "server_ip": "10.1.0.101"}, 30, 1, "bandwidth_link_usage_post_failure", "TCP Bandwidth after failure", "tcp", 100,
                    [0, 130]))
                },
                "failures": [
                    {
                        "type": "intermediate",
                        "execute": {
                            "use_pre_defined_function": True,
                            "command": (
                                "connection_shutdown", (['r2', 'r4'], ['Router 2', 'Router 4'], ['r2-eth2', 'r4-eth1']))
                        }
                    }
                ]
            },
            "bandwidth_link_usage_concurrent": {
                "pre_execution": {
                    "use_pre_defined_function": False,
                    "command": lambda net: net.iperf((net['h1'], net['h4'])),
                },
                "source": "h1",
                "destination": "h4",
                "destination_ip": "10.4.0.101",
                "execute": {
                    "use_pre_defined_function": True,
                    "command": ("measure_link_usage_bandwidth", (
                        {"client": "h1", "server": "h4", "server_ip": "10.4.0.101"},
                        {"client": "h2", "server": "h1", "server_ip": "10.1.0.101"}, 30, 1,
                        "bandwidth_link_usage_concurrent_failure", "TCP Bandwidth with concurrent failure", "tcp", 100,
                        [0, 130])),
                },
                "failures": [
                    {
                        "type": "timer",
                        "timing": 18,
                        "execute": {
                            "use_pre_defined_function": True,
                            "command": (
                                "connection_shutdown", (['r2', 'r4'], ['Router 2', 'Router 4'], ['r2-eth2', 'r4-eth1']))
                        }
                    }
                ]
            },
            "bandwidth_failure_concurrent": {
                "pre_execution": {
                    "use_pre_defined_function": False,
                    "command": lambda net: net.iperf((net['h1'], net['h4'])),
                },
                "source": "h1",
                "destination": "h4",
                "destination_ip": "10.4.0.101",
                "execute": {
                    "use_pre_defined_function": True,
                    "command": ("measure_bandwidth", (
                    ['h1', 'h4'], '10.4.0.101', 30, 0.5, "tcp_failure_concurrent", "TCP Bandwidth with concurrent failure",
                    "tcp", 100, [0, self.limit + self.limit * 0.1])),

                },
                "failures": [
                    {
                        "type": "timer",
                        "timing": 18,
                        "execute": {
                            "use_pre_defined_function": True,
                            "command": (
                                "connection_shutdown", (['r2', 'r4'], ['Router 2', 'Router 4'], ['r2-eth2', 'r4-eth1']))
                        }
                    }
                ]
            },
            "latency_failure_concurrent": {
                "pre_execution": {
                    "use_pre_defined_function": False,
                    "command": lambda net: net.ping((net['h1'], net['h4'])),
                },
                "source": "h1",
                "destination": "h4",
                "destination_ip": "10.4.0.101",
                "execute": {
                    "use_pre_defined_function": True,
                    "command": ("measure_latency", (
                    'h1', '10.4.0.101', 30, 0.5, "s_to_d_latency_concurrent", [0, 130], "Latency from H1 to H4 with concurrent failure"))

                },
                "failures": [
                    {
                        "type": "timer",
                        "timing": 15,
                        "execute": {
                            "use_pre_defined_function": True,
                            "command": (
                                "connection_shutdown", (['r2', 'r4'], ['Router 2', 'Router 4'], ['r2-eth2', 'r4-eth1']))
                        }
                    }
                ]
            },
            "latency_failure_intermediate": {
                "pre_execution": {
                    "use_pre_defined_function": False,
                    "command": lambda net: net.ping((net['h1'], net['h4'])),
                },
                "source": "h1",
                "destination": "h4",
                "destination_ip": "10.4.0.101",
                "execute": {
                    "use_pre_defined_function": True,
                    "separate_definitions": True,
                    "command_pre": ("measure_latency", (
                        'h1', '10.4.0.101', 30, 0.5, "before_failure", [0, 100], "Latency from H1 to H4 before failure")),
                    "command_post": ("measure_latency", (
                        'h1', '10.4.0.101', 30, 0.5, "after_failure", [0, 100], "Latency from H1 to H4 after failure"))

                },
                "failures": [
                    {
                        "type": "intermediate",
                        "execute": {
                            "use_pre_defined_function": True,
                            "command": (
                                "connection_shutdown", (['r2', 'r4'], ['Router 2', 'Router 4'], ['r2-eth2', 'r4-eth1']))
                        }
                    }
                ]
            },
            "tcp_packet_flow_intermediate": {
                "pre_execution": {
                    "use_pre_defined_function": False,
                    "command": lambda net: net.iperf((net['h1'], net['h4'])),
                },
                "source": "h1",
                "destination": "h4",
                "destination_ip": "10.4.0.101",
                "execute": {
                    "use_pre_defined_function": True,
                    "separate_definitions": True,
                    "command_pre": ("measure_packet_flow", (
                        'h1', 'h4', '10.4.0.101', ["r1", "r2", "r3", "r4"], 30, 1, "before_failure", [0, 3000], "Packet flow on all routers before failure", "tcp", 100)),
                    "command_post": ("measure_packet_flow", (
                        'h1', 'h4', '10.4.0.101', ["r1", "r2", "r3", "r4"], 30, 1, "after_failure", [0, 3000], "Packet flow on all routers after failure", "tcp", 100)),

                },
                "failures": [
                    {
                        "type": "intermediate",
                        "execute": {
                            "use_pre_defined_function": True,
                            "command": (
                                "connection_shutdown", (['r2', 'r4'], ['Router 2', 'Router 4'], ['r2-eth2', 'r4-eth1']))
                        }
                    }
                ]
            },
            "udp_packet_flow_intermediate": {
                "pre_execution": {
                    "use_pre_defined_function": False,
                    "command": lambda net: net.iperf((net['h1'], net['h4'])),
                },
                "source": "h1",
                "destination": "h4",
                "destination_ip": "10.4.0.101",
                "execute": {
                    "use_pre_defined_function": True,
                    "separate_definitions": True,
                    "command_pre": ("measure_packet_flow", (
                        'h1', 'h4', '10.4.0.101', ["r1", "r2", "r3", "r4"], 30, 1, "udp_before_failure", [0, 22000],
                        "UDP Packet flow on all routers before failure", "udp", 100)),
                    "command_post": ("measure_packet_flow", (
                        'h1', 'h4', '10.4.0.101', ["r1", "r2", "r3", "r4"], 30, 1, "udp_after_failure", [0, 22000],
                        "UDP Packet flow on all routers after failure", "udp", 100)),

                },
                "failures": [
                    {
                        "type": "intermediate",
                        "execute": {
                            "use_pre_defined_function": True,
                            "command": (
                                "connection_shutdown", (['r2', 'r4'], ['Router 2', 'Router 4'], ['r2-eth2', 'r4-eth1']))
                        }
                    }
                ]
            },
            "udp_packet_flow_concurrent": {
                "pre_execution": {
                    "use_pre_defined_function": False,
                    "command": lambda net: net.iperf((net['h1'], net['h4'])),
                },
                "source": "h1",
                "destination": "h4",
                "destination_ip": "10.4.0.101",
                "execute": {
                    "use_pre_defined_function": True,
                    "command": ("measure_packet_flow", (
                        'h1', 'h4', '10.4.0.101', ["r1", "r2", "r3", "r4"], 30, 1, "udp_concurrent_failure", [0, 22000],
                        "UDP Packet flow on all routers before failure", "udp", 100)),

                },
                "failures": [
                    {
                        "type": "timer",
                        "timing": 19.1,
                        "execute": {
                            "use_pre_defined_function": True,
                            "command": (
                                "connection_shutdown", (['r2', 'r4'], ['Router 2', 'Router 4'], ['r2-eth2', 'r4-eth1']))
                        }
                    }
                ]
            },
            "tcp_packet_flow_concurrent": {
                "pre_execution": {
                    "use_pre_defined_function": False,
                    "command": lambda net: net.iperf((net['h1'], net['h4'])),
                },
                "source": "h1",
                "destination": "h4",
                "destination_ip": "10.4.0.101",
                "execute": {
                    "use_pre_defined_function": True,
                    "command": ("measure_packet_flow", (
                        'h1', 'h4', '10.4.0.101', ["r1", "r2", "r3", "r4"], 30, 1, "concurrent_failure", [0, 3000], "Packet flow on all routers before failure", "tcp", 100)),

                },
                "failures": [
                    {
                        "type": "timer",
                        "timing": 19.1,
                        "execute": {
                            "use_pre_defined_function": True,
                            "command": (
                                "connection_shutdown", (['r2', 'r4'], ['Router 2', 'Router 4'], ['r2-eth2', 'r4-eth1']))
                        }
                    }
                ]
            }

        }
